/* Example Arduino program for setting up and controlling an EasyDAQ I2C/SPI-4VI4xx I2C/SPI controlled Relay and input card to be controlled by I2C.

This example sends buye values from a serial source set to 115200 baud, 8Bits 1 stop bit and no parity via an Arduino to the I2C/SPI-4VI4xx relays.

Configuration of the SC16IS750 Baud rate generator to 9600 baud (1.8432MHz crystal option), sending Port B data.

Data is sent back via the Arduino USB to the serial port which should be set to 115200 baud, 8 Bits 1 stop bit and no parity.

This program has a "debug" element which converts returning values to readable, hex numbers.

For further information please contact EasyDAQ

NJH 17/09/21
*/

#include <Wire.h>
byte Address = 0x48;          // Address of the SC16IS750 when A0 and A1 are pulled high.
                              // The value in the datasheet has to be shifted right by 1 bit to allow for the R/W bit.
int delaytime = 25;           // 25ms min delay between instructions

// Register addresses shifted left 3 bits into bits 3:6. See NXP SC16IS740/750/760 datasheet Rev. 7 — 9 June 2011 section 10.4 "Use of subaddresses"

// SC16IS750 Registers :         Only a few of these are actually used in this program.
#define THR        0x00 << 3  // Transmit Holding Register. Can only be written to.      
#define RHR        0x00 << 3  // Receive Holding Register. Can only be read from.     
#define IER        0x01 << 3  // Interrupt Enable Register
#define FCR        0x02 << 3  // FIFO Control Register
#define IIR        0x02 << 3  // Modem Control Register
#define LCR        0x03 << 3  // Line Control Register
#define MCR        0x04 << 3  // Modem Control register
#define LSR        0x05 << 3  // Line Status Register
#define MSR        0x06 << 3  // Modem Status Register
#define SPR        0x07 << 3  // Scratch Pad Register
#define TXLVL      0x08 << 3  // Transmitter FIFO Level Register
#define RXLVL      0x09 << 3  // Receiver FIFO level
#define IODIR      0x0A << 3  // I/O pin Direction Register
#define IOSTATE    0x0B << 3  // I/O pin States Register
#define IOINTMSK   0x0C << 3  // I/O Interrupt Enable Register
#define IOControl  0x0e << 3  // I/O Control Register
#define EFCR       0x0f << 3  // Extra Features Control Register

#define DLL        0x00 << 3  // divisor latch LSB
#define DLH        0x01 << 3  // divisor latch MSB
#define EFR        0x02 << 3  // Enhanced Features Register
#define XON1       0x04 << 3  // Xon1 word
#define XON2       0x05 << 3  // Xon2 word
#define XOFF1      0x06 << 3  // Xoff1 word
#define XOFF2      0x07 << 3  // Xoff2 word

int ledState = 0;
byte outgoingByte = 0;
byte incomingByte = 0;  // for incoming serial data
byte debug = true;      // set to false if the debug informaion is not required

void setup() {
  Wire.begin();                       // join i2c bus (address optional for master)
  if(debug){
    Serial.begin(115200);               // start serial for debug output
    for (int i = 0; i <= 9; i++){       // clear a bit of the serial monitor view
      Serial.println("");
    }
    Serial.println("EasyDAQ Serial to I2C program");
    Serial.println("=============================");
  }
  setupBRG();                         // Set up the SC16IS750 on the EasyDAQ card to talk to the processor
  setupI2C4VI4();                     // Setup the EasyDAQ card processor
}

void loop(){
  if(Serial.available() > 0){
    incomingByte = Serial.read();
    writeBytePortB(incomingByte);
    delay(delaytime);
  }
  if(getByte(RXLVL) > 0){
    outgoingByte = getByte(RHR);      // Get the data
    Serial.write(outgoingByte);
    if(debug) {
      Serial.print(" Debug: Returned value = 0x");
      if(outgoingByte < 16) Serial.print("0");
      Serial.println(outgoingByte, HEX);
    }
  }
}

void setupBRG(){
  Serial.println("setting up BRG");
  putByte(IOControl,B00001000);       // Software reset
  while ((getByte(IOControl) & B00001000) > 0){ // Check it has reset
    if(debug){
      Serial.println("Waiting");
    }
  }
  putByte(LCR,0x80);                  // Divisor latch enable
                                      // Set up BRG to 9600 baud
  putByte(DLL,12);                    // Lower byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 
  putByte(DLH,0x00);                  // Upper byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 

  putByte(LCR,0xBF);                  // Access EFR register
  putByte(EFR,0x10);                  // Enable Enhanced functions
  putByte(LCR,0x03);                  // Setup LCR for normal operation: 8 bit data, 1 stop bits, no parity
  putByte(FCR,0x06);                  // Reset FIFOs
  putByte(FCR,0x01);                  // Enable FIFOs
  if(debug){
    Serial.println("BRG set up");
  }
}

byte getByte(byte reg){               // Receives data from the SC16IS750
  byte c;
  Wire.beginTransmission(Address);    // transmit to device
  Wire.write(reg);                    // Send the address
  Wire.endTransmission();             // stop I2C transmitting
  
  Wire.requestFrom(Address, 1);       // Request 1 byte from slave device. I.e. the current value in the Receive Holding Register (RHR)
                                      // the I2C buffer will fill with the requested number of bytes
  while (Wire.available()) {          // slave may send More than requested
    c = Wire.read();                  // receive a byte from the buffer
  }
  return c;
}

void putByte(byte reg, byte data){    // Sends I2C data to the SC16IS750
  Wire.beginTransmission(Address);    // transmit to device
  Wire.write(reg);                    // Send the register address
  Wire.write(data);                   // Send the data
  Wire.endTransmission();             // stop transmitting
}

void setupI2C4VI4(){
  putByte(THR, 0x42);                 // Set direction of port B to lower 4 bits as output, upper 4 bits as input
  putByte(THR, 0xF0);
  delay(delaytime);
  Serial.println("EasyDAQ I2C4VI4xx setup");
  Serial.println();
}

byte readPortB(){
  Serial.println("Reading port B");
  byte RxFIFO = 0;
  putByte(THR, 0x41);                 // Request a byte from Port B
  putByte(THR, 0x00);                 // filler byte
  unsigned long timeout = millis();   // use a timeout to ensure that the function will return a result.
  //while ((RxFIFO == 0) && ((millis() - timeout) < 1000)){  // wait for the byte to arrive
  while (RxFIFO == 0){
    RxFIFO = getByte(RXLVL);          // Has a byte arrived in the FIFO?
  }
  if ((millis() - timeout) >= 1000) Serial.println("Read port B timeout");
  Serial.print("RXLVL = ");
  Serial.println(RxFIFO);
  byte b = getByte(RHR);              // Get the required data
  return b;
}

void writeBytePortB(byte dataout){
  putByte(THR, dataout);              // Data byte
}
