/* Example Arduino program for setting up and controlling an EasyDAQ I2C/SPI-4VI4xx I2C/SPI controlled Relay and input card to be controlled by I2C.

This example reads port B and controls all the relays on port B dependaing on the value of input 1 (Bit 4 of port B, In 1 of the 4 wide inputs)

Configuration of the SC16IS750 Baud rate generator to 9600 baud (1.8432MHz crystal option), sending and receiving Port B data.

Debug data is sent back via the Arduino USB to the "Serial monitor" which should be set to 115200 for this example.

If the COM and NO outputs of one of the relays is used to control power to input 1 then the circuit will switch all the relays on and off at maximum speed.

For further information please contact EasyDAQ

NJH 13/12/19
*/

#include <Wire.h>
byte Address = 0x48;          // Address of the SC16IS750 when A0 and A1 are pulled high.
                              // The value in the datasheet has to be shifted right by 1 bit to allow for the R/W bit.
int delaytime = 25;           // 25ms min delay between instructions

// Register addresses shifted left 3 bits into bits 3:6. See NXP SC16IS740/750/760 datasheet Rev. 7 — 9 June 2011 section 10.4 "Use of subaddresses"

// SC16IS750 Registers :         Only a few of these are actually used in this program.
#define THR        0x00 << 3  // Transmit Holding Register. Can only be written to.      
#define RHR        0x00 << 3  // Receive Holding Register. Can only be read from.     
#define IER        0x01 << 3  // Interrupt Enable Register
#define FCR        0x02 << 3  // FIFO Control Register
#define IIR        0x02 << 3  // Modem Control Register
#define LCR        0x03 << 3  // Line Control Register
#define MCR        0x04 << 3  // Modem Control register
#define LSR        0x05 << 3  // Line Status Register
#define MSR        0x06 << 3  // Modem Status Register
#define SPR        0x07 << 3  // Scratch Pad Register
#define TXLVL      0x08 << 3  // Transmitter FIFO Level Register
#define RXLVL      0x09 << 3  // Receiver FIFO level
#define IODIR      0x0A << 3  // I/O pin Direction Register
#define IOSTATE    0x0B << 3  // I/O pin States Register
#define IOINTMSK   0x0C << 3  // I/O Interrupt Enable Register
#define IOControl  0x0e << 3  // I/O Control Register
#define EFCR       0x0f << 3  // Extra Features Control Register

#define DLL        0x00 << 3  // divisor latch LSB
#define DLH        0x01 << 3  // divisor latch MSB
#define EFR        0x02 << 3  // Enhanced Features Register
#define XON1       0x04 << 3  // Xon1 word
#define XON2       0x05 << 3  // Xon2 word
#define XOFF1      0x06 << 3  // Xoff1 word
#define XOFF2      0x07 << 3  // Xoff2 word

int ledState = 0;

void setup() {                        // Setup.
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, !ledState);
  ledState = !ledState;
  Wire.begin();                       // join i2c bus (address optional for master)
  Serial.begin(115200);               // start serial for debug output
  for (int i = 0; i <= 9; i++){       // clear a bit of the serial monitor view
    Serial.println("");
  }
  setupBRG();                         // Set up the SC16IS750 on the EasyDAQ card to talk to the processor
  setupI2C4VI4();                     // Setup the EasyDAQ card processor
}

void loop(){                          // Main program loop
  digitalWrite(LED_BUILTIN, !ledState); // Loop running indicator on the built in LED
  ledState = !ledState;
  Serial.println("Beginning of loop");
  byte data = readPortB();
  delay(delaytime);
  Serial.print("Data from Port B = ");
  Serial.println(data, HEX);
  if ((data & 0x10) == 0){            // switch all the relays according to the value of input 1 (Port B bit 4)
    data = 0xff;
  }
  else data = 0;
  Serial.print("Data to Port B   = ");
  Serial.println(data, HEX);
  writePortB(data);
  Serial.println();
  delay(delaytime);
  //delay(1000);                      // enable this line to slow the loop to 1 sample per second
}

void setupBRG(){
  Serial.println("setting up BRG");
  putByte(IOControl,B00001000);       // Software reset
  while ((getByte(IOControl) & B00001000) > 0){ // Check it has reset
    Serial.println("Waiting");
  }
  putByte(LCR,0x80);                  // Divisor latch enable
                                      // Set up BRG to 9600 baud
  putByte(DLL,12);                    // Lower byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 
  putByte(DLH,0x00);                  // Upper byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 

  putByte(LCR,0xBF);                  // Access EFR register
  putByte(EFR,0x10);                  // Enable Enhanced functions
  putByte(LCR,0x03);                  // Setup LCR for normal operation: 8 bit data, 1 stop bits, no parity
  putByte(FCR,0x06);                  // Reset FIFOs
  putByte(FCR,0x01);                  // Enable FIFOs
  Serial.println("BRG set up");
  Serial.println("");
}

byte getByte(byte reg){               // Receives data from the SC16IS750
  byte c;
  Wire.beginTransmission(Address);    // transmit to device
  Wire.write(reg);                    // Send the address
  Wire.endTransmission();             // stop I2C transmitting
  
  Wire.requestFrom(Address, 1);       // Request 1 byte from slave device. I.e. the current value in the Receive Holding Register (RHR)
                                      // the I2C buffer will fill with the requested number of bytes
  while (Wire.available()) {          // slave may send More than requested
    c = Wire.read();                  // receive a byte from the buffer
  }
  return c;
}

void putByte(byte reg, byte data){    // Sends I2C data to the SC16IS750
  Wire.beginTransmission(Address);    // transmit to device
  Wire.write(reg);                    // Send the register address
  Wire.write(data);                   // Send the data
  Wire.endTransmission();             // stop transmitting
}

void setupI2C4VI4(){
  putByte(THR, 0x42);                 // Set direction of port B to lower 4 bits as output for the 4 relays, upper 4 bits as input
  putByte(THR, 0xF0);
  delay(delaytime);
}

byte readPortB(){                     // Get 1 byte of data from Port B. Top 4 bits will be In 1 to In 4 and the bottom 4 bits will be the relay states.
  byte RxFIFO = 0;
  putByte(THR, 0x41);                 // Request a byte from Port B
  putByte(THR, 0x00);                 // filler byte
  unsigned long timeout = millis();   // use a timeout to ensure that the function will return a result.
  while ((RxFIFO == 0) && ((millis() - timeout) < 1000)){  // wait for the byte to arrive
    RxFIFO = getByte(RXLVL);          // Has a byte arrived in the FIFO?
  }
  if ((millis() - timeout) >= 1000) Serial.println("Read port B timeout");
  byte b = getByte(RHR);              // Get the required data
  return b;
}

void writePortB(byte dataout){
  putByte(THR, 0x43);                 // Send a byte to Port B
  putByte(THR, dataout);              // Data byte
}
