/* Example Arduino sketch for setting up and controlling an EasyDAQ I2C/SPI4VI4xx I2C/SPI controlled Relay and input card to be controlled by SPI.

This example reads port B and controls all the relays on port B dependaing on the value of input 1 (Bit 4)

Configuration of the SC16IS750 Baud rate generator to 9600 baud (1.8432MHz crystal option), sending Port B data and receiving port C data.

Debug data is sent back via the Arduino USB to the "Serial monitor" which should be set to 115200 for this example.

If the COM and NO outputs of one of the relays is used to control power to input 1 then the circuit will switch all the relays on and off at maximum speed.

See "https://www.arduino.cc/en/Tutorial/SPITransaction" for the use of Arduino SPI transaction methods.

For further information please contact EasyDAQ

NJH 13/12/19
*/

#include <SPI.h>
const int ss = 10;
const unsigned long speedMaximum = 1000000; // 4000000 = 4MHz maximum SPI data rate of SC16IS750/750/760. Adjust for reliability in your application.
SPISettings SC16IS750SPISetttings(speedMaximum, MSBFIRST, SPI_MODE0);

int delaytime = 25;           // 25ms min delay between instructions

// Register addresses shifted left 3 bits into bits 3:6. See NXP SC16IS750/750/760 datasheet Rev. 7 — 9 June 2011 section 10.4 "Use of subaddresses"
// 
// SC16IS750 Registers :         Only a few of these are actually used in this program.
#define THR        0x00 << 3  // Transmit Holding Register. Can only be written to.      
#define RHR        0x00 << 3  // Receive Holding Register. Can only be read from.     
#define IER        0x01 << 3  // Interrupt Enable Register
#define FCR        0x02 << 3  // FIFO Control Register
#define IIR        0x02 << 3  // Modem Control Register
#define LCR        0x03 << 3  // Line Control Register
#define MCR        0x04 << 3  // Modem Control register
#define LSR        0x05 << 3  // Line Status Register
#define MSR        0x06 << 3  // Modem Status Register
#define SPR        0x07 << 3  // Scratch Pad Register
#define TXLVL      0x08 << 3  // Transmitter FIFO Level Register
#define RXLVL      0x09 << 3  // Receiver FIFO level
#define IODIR      0x0A << 3  // I/O pin Direction Register
#define IOSTATE    0x0B << 3  // I/O pin States Register
#define IOINTMSK   0x0C << 3  // I/O Interrupt Enable Register
#define IOControl  0x0e << 3  // I/O Control Register
#define EFCR       0x0f << 3  // Extra Features Control Register

#define DLL        0x00 << 3  // divisor latch LSB
#define DLH        0x01 << 3  // divisor latch MSB
#define EFR        0x02 << 3  // Enhanced Features Register
#define XON1       0x04 << 3  // Xon1 word
#define XON2       0x05 << 3  // Xon2 word
#define XOFF1      0x06 << 3  // Xoff1 word
#define XOFF2      0x07 << 3  // Xoff2 word
  
void setup() {
  pinMode(ss, OUTPUT);
  digitalWrite(ss, HIGH);             // take the chip select high to de-select
  SPI.begin();
  Serial.begin(115200);               // start serial for debug output
  for (int i = 0; i <= 9; i++){       // clear a bit of the serial monitor view
    Serial.println("");
  }
  setupBRG();                         // Set up the SC16IS750 on the EasyDAQ card to talk to the processor
  setupSPI4VI4();                     // Setup the EasyDAQ card processor
}

void loop(){
  byte data = readPortB();
  delay(delaytime);
  Serial.print("Data from Port B = ");
  Serial.println(data, HEX);
  if ((data & 0x10) == 0){            // switch all the relays according to the value of input 1
    data = 0xff;
  }
  else data = 0;
  Serial.print("  Data to Port B = ");
  Serial.println(data, HEX);
  writePortB(data);
  delay(delaytime);
}

void setupBRG(){
  Serial.println("setting up BRG");
  putByte(IOControl,B00001000);       // Software reset
  while ((getByte(IOControl) & B00001000) > 0){ // Check it has reset
    Serial.println("Waiting");
  }
  putByte(LCR,0x80);                  // Divisor latch enable
                                      // Set up BRG to 9600 baud
  putByte(DLL,12);                    // Lower byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 
  putByte(DLH,0x00);                  // Upper byte of BRG  ((1.8432MHz/(12x16)). 16 clocks required per bit sent 

  putByte(LCR,0xBF);                  // Access EFR register
  putByte(EFR,0x10);                  // Enable Enhanced functions
  putByte(LCR,0x03);                  // Setup LCR for normal operation: 8 bit data, 1 stop bits, no parity
  putByte(FCR,0x06);                  // Reset FIFOs
  putByte(FCR,0x01);                  // Enable FIFOs
  Serial.println("BRG set up");
  Serial.println("");
}

byte getByte(byte reg){               // Receives data from the SC16IS750
  byte c;
  SPI.beginTransaction(SC16IS750SPISetttings);
  digitalWrite(ss,LOW);
  SPI.transfer(reg | 0x80);           // Send the register address, add read mode by setting the top bit
  c = SPI.transfer(0xFF);             // get data byte, a dummy byte is sent to clock out the byte
  digitalWrite(ss, HIGH);             // take the chip select high to de-select:
  SPI.endTransaction();
  return c;
}

void putByte(byte reg, byte data){    // Sends SPI data to the SC16IS750
  SPI.beginTransaction(SC16IS750SPISetttings);
  digitalWrite(ss, LOW);              // take the chip select low to select the device:
  SPI.transfer(reg);                  // Send the register address
  SPI.transfer(data);                 // Send the data
  digitalWrite(ss, HIGH);             // take the chip select high to de-select:
  SPI.endTransaction();
}

void setupSPI4VI4(){
  putByte(THR, 0x42);                 // Set direction of lower 4 bits of port B to output for relays and upper 4 bits for inputs
  putByte(THR, 0xF0);
  delay(delaytime);
}

byte readPortB(){                     // Get 1 byte of data from Port B. Top 4 bits will be In 1 to In 4 and the bottom 4 bits will be the relay states.
  putByte(THR, 0x41);                 // Request a byte from Port B
  putByte(THR, 0x00);                 // filler byte
  byte RxFIFO = 0;
  unsigned long timeout = millis();   // use a timeout to ensure that the function will return a result.
  while ((RxFIFO == 0) && ((millis() - timeout) < 1000)){  // wait for the byte to arrive
    RxFIFO = getByte(RXLVL);          // Has a byte arrived in the FIFO?
  }
  if ((millis() - timeout) >= 1000) Serial.println("Read port B timeout");
  byte b = getByte(RHR);              // Get the required data
  return b;
}

void writePortB(byte dataout){
  putByte(THR, 0x43);                 // Send a byte to Port B
  putByte(THR, dataout);              // Data byte
}
